"""This module gathers misc convenience functions to handle s3 objects"""
import logging
import re
import tempfile
from time import sleep
from typing import IO, Any, Dict, Optional, Tuple, cast
from urllib.parse import unquote, urlparse

import s3fs

S3_SCHEMES = ["s3", "s3n", "s3a"]

logger = logging.getLogger(__name__)


def parse_s3_url(
    url: str, file: bool = True
) -> Tuple[Optional[str], Optional[str], Optional[str], str]:
    """parses a s3 url and extract credentials and s3 object path.

    A S3 URL looks like s3://aws_key:aws_secret@bucketname/objectname where
    credentials are optional. Since credentials might include characters
    such as `/`, `@` or `#`, they have to be urlquoted in the url.

    Args:
        url (str): the s3 url
        file (bool): whether or not the url is a file url or a directory one

    Returns:
        tuple: (access_key, secret, bucketname, objectname). If credentials
        are not specified, `access_key` and `secret` are set to None.
    """
    urlchunks = urlparse(url)
    scheme = urlchunks.scheme
    assert scheme in S3_SCHEMES, f"{scheme} unsupported, use one of {S3_SCHEMES}"
    assert not urlchunks.params, f"s3 url should not have params, got {urlchunks.params}"
    assert not urlchunks.query, f"s3 url should not have query, got {urlchunks.query}"
    assert not urlchunks.fragment, f"s3 url should not have fragment, got {urlchunks.fragment}"

    access_key: Optional[str] = None
    secret: Optional[str] = None

    # if either username or password is specified, we have credentials
    if urlchunks.username or urlchunks.password:
        # and they should both not be empty
        assert urlchunks.username, "s3 access key should not be empty"
        assert urlchunks.password, "s3 secret should not be empty"
        access_key = unquote(urlchunks.username)
        secret = unquote(urlchunks.password)
    objectname = urlchunks.path.lstrip("/")  # remove leading /, it's not part of the objectname
    if file:
        assert objectname, "s3 objectname can't be empty"

    return access_key, secret, urlchunks.hostname, objectname


def _s3_open_file_with_retries(fs: s3fs.S3FileSystem, path: str, retries: int) -> Any:
    for _ in range(retries):
        try:
            logger.info(f"opening {path}")
            file = fs.open(path)
            return file
        except Exception as ex:
            logger.warning(f"could not open {path}: {ex}")
            # if the file has just been uploaded, then it might not be visible immediatly
            # but the fail to open has been cached by s3fs
            # so, we invalidate the cache
            fs.invalidate_cache(path)
            # and we give some time to S3 to settle the file status
            sleep(1)


def s3_open(url: str, *, client_kwargs: Optional[Dict[str, Any]] = None) -> IO[bytes]:
    """opens a s3 url and returns a file-like object"""
    access_key, secret, bucketname, objectname = parse_s3_url(url)
    fs = s3fs.S3FileSystem(key=access_key, secret=secret, client_kwargs=client_kwargs)
    path = f"{bucketname}/{objectname}"
    ret = tempfile.NamedTemporaryFile(suffix=".s3tmp")
    file = _s3_open_file_with_retries(fs, path, 3)
    ret.write(file.read())
    ret.seek(0)
    file.close()
    return ret


def _get_timestamp(obj: Dict[str, Any]) -> Optional[int]:
    try:
        return cast(int, obj["LastModified"].timestamp())
    except KeyError:
        return None


def s3_mtime(url: str, *, client_kwargs: Optional[Dict[str, Any]] = None) -> Optional[int]:
    access_key, secret, bucketname, objectname = parse_s3_url(url, file=True)
    fs = s3fs.S3FileSystem(key=access_key, secret=secret, client_kwargs=client_kwargs)
    return _get_timestamp(fs.info(f"{bucketname}/{objectname}"))


def dir_mtimes(
    dirpath: str, *, client_kwargs: Optional[Dict[str, Any]] = None
) -> Dict[str, Optional[int]]:
    access_key, secret, bucketname, objectname = parse_s3_url(dirpath, file=False)
    fs = s3fs.S3FileSystem(key=access_key, secret=secret, client_kwargs=client_kwargs)
    # objectname can be empty or not ('subdir1/subdir2')
    bucketdir = f"{bucketname}/{objectname}".rstrip("/")
    return {
        re.sub(rf"^{bucketdir}/", "", x["name"]): _get_timestamp(x) for x in fs.listdir(bucketdir)
    }



from abc import ABCMeta, abstractmethod
from contextlib import suppress
from datetime import timedelta
from enum import Enum
from pathlib import Path
from time import time
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

import pandas as pd

if TYPE_CHECKING:
    from typing_extensions import TypedDict

    class InMemoryCached(TypedDict):
        value: pd.DataFrame
        mtime: float
        created_at: float


class CacheEnum(str, Enum):
    MEMORY = "memory"
    HDF = "hdf"


class Cache(metaclass=ABCMeta):
    @staticmethod
    def get_cache(kind: CacheEnum, *args: Any, **kwargs: Any) -> "Cache":
        ALL_CACHES = {
            CacheEnum.MEMORY: InMemoryCache,
            CacheEnum.HDF: HDFCache,
        }
        return ALL_CACHES[kind](*args, **kwargs)  # type: ignore[no-any-return]

    @staticmethod
    def should_invalidate(
        *,
        mtime: Optional[float] = None,
        cached_mtime: float,
        expire: Optional[timedelta] = None,
        cached_created_at: float,
    ) -> bool:
        now = time()
        is_newer_version = False
        is_expired = False
        if mtime is not None:
            is_newer_version = mtime != cached_mtime
        if expire is not None:
            is_expired = now > cached_created_at + expire.total_seconds()
        return is_newer_version or is_expired

    @abstractmethod
    def get(
        self, key: str, mtime: Optional[float] = None, expire: Optional[timedelta] = None
    ) -> pd.DataFrame:
        """get a cached value"""

    @abstractmethod
    def set(self, key: str, value: pd.DataFrame, mtime: Optional[float] = None) -> None:
        """set a cached value"""

    @abstractmethod
    def delete(self, key: str) -> None:
        """delete a cached value"""


class InMemoryCache(Cache):
    def __init__(self) -> None:
        self._cache: Dict[str, "InMemoryCached"] = {}

    def get(
        self, key: str, mtime: Optional[float] = None, expire: Optional[timedelta] = None
    ) -> pd.DataFrame:
        cached = self._cache[key]
        if self.should_invalidate(
            mtime=mtime,
            cached_mtime=cached["mtime"],
            expire=expire,
            cached_created_at=cached["created_at"],
        ):
            self.delete(key)
        return self._cache[key]["value"]

    def set(self, key: str, value: pd.DataFrame, mtime: Optional[float] = None) -> None:
        mtime = mtime or time()
        self._cache[key] = {"value": value, "mtime": mtime, "created_at": time()}

    def delete(self, key: str) -> None:
        if key in self._cache:
            del self._cache[key]


class HDFCache(Cache):
    META_DF_KEY = "__meta__"

    def __init__(self, cache_dir: Union[str, Path]) -> None:
        self.cache_dir = Path(cache_dir).resolve()

    def get_metadata(self) -> pd.DataFrame:
        """
        metadata is a dataframe containing last mtime and created_at fields
        for each cached datasource, identified by its key (= its hash).
        If metadata file is not found or is corrupted, an empty one is recreated.
        """
        try:
            # We manually instantiate the HDFStore to be able to close it no matter what
            # See https://github.com/pandas-dev/pandas/pull/28429 for more infos
            store = pd.HDFStore(self.cache_dir / self.META_DF_KEY, mode="r")
            try:
                metadata = pd.read_hdf(store)
            finally:
                store.close()
        except Exception:  # catch all, on purpose
            metadata = pd.DataFrame(columns=["key", "mtime", "created_at"])
            self.set_metadata(metadata)
        return metadata

    def set_metadata(self, df: pd.DataFrame) -> None:
        df.to_hdf(self.cache_dir / self.META_DF_KEY, self.META_DF_KEY, mode="w")

    def get(
        self, key: str, mtime: Optional[float] = None, expire: Optional[timedelta] = None
    ) -> pd.DataFrame:
        metadata = self.get_metadata()
        try:
            # look for the row concerning the desired key in the metadata dataframe:
            infos = metadata[metadata.key == key].iloc[0].to_dict()
        except IndexError:
            raise KeyError(key)

        if self.should_invalidate(
            mtime=mtime,
            cached_mtime=infos["mtime"],
            expire=expire,
            cached_created_at=infos["created_at"],
        ):
            self.delete(key)

        try:
            return pd.read_hdf(self.cache_dir / key)
        except FileNotFoundError:
            raise KeyError(key)

    def set(self, key: str, value: pd.DataFrame, mtime: Optional[float] = None) -> None:
        mtime = mtime or time()
        infos = {"key": key, "mtime": mtime, "created_at": time()}
        metadata = self.get_metadata()
        try:
            # add new row to the metadata dataframe:
            metadata = metadata[metadata.key != key]  # drop duplicates
            metadata = pd.concat([metadata, pd.Series(infos).to_frame().T], ignore_index=True)
            self.set_metadata(metadata)
            value.to_hdf(self.cache_dir / key, key, mode="w")
        except OSError:
            self.delete(key)
            raise

    def delete(self, key: str) -> None:
        metadata = self.get_metadata()
        metadata = metadata[metadata.key != key]
        self.set_metadata(metadata)
        with suppress(FileNotFoundError):
            (self.cache_dir / key).unlink()



import os
from typing import IO, Any, Dict, List, Optional

from ..fetcher import Fetcher, register
from .ftp_utils import FTP_SCHEMES, dir_mtimes, ftp_mtime, ftp_open


@register(schemes=FTP_SCHEMES)
class FTPFetcher(Fetcher):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._mtimes_cache: Dict[str, Dict[str, Optional[int]]] = {}

    def get_dir_mtimes(self, dirpath: str) -> Dict[str, Optional[int]]:
        if dirpath not in self._mtimes_cache:
            self._mtimes_cache[dirpath] = dir_mtimes(dirpath)
        return self._mtimes_cache[dirpath]

    def open(self, filepath: str) -> IO[bytes]:
        return ftp_open(filepath)

    def listdir(self, dirpath: str) -> List[str]:
        """
        Make use of listdir to get all mtimes of remote files at once and
        put this information in cache.
        This will save us time by avoiding to create a new FTP connection
        for each file when we'll want to get its mtime.
        """
        return list(self.get_dir_mtimes(dirpath).keys())

    def mtime(self, filepath: str) -> Optional[int]:
        dirpath, filename = os.path.split(filepath)
        if dirpath in self._mtimes_cache:
            return self._mtimes_cache[dirpath][filename]
        else:
            return ftp_mtime(filepath)



import os
from typing import IO, Any, Dict, List, Optional

from ..fetcher import Fetcher, register
from .s3_utils import S3_SCHEMES, dir_mtimes, s3_mtime, s3_open


@register(schemes=S3_SCHEMES)
class S3Fetcher(Fetcher):
    def __init__(self, *, client_kwargs: Optional[Dict[str, Any]] = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.client_kwargs = client_kwargs
        self._mtimes_cache: Dict[str, Dict[str, Optional[int]]] = {}

    def get_dir_mtimes(self, dirpath: str) -> Dict[str, Optional[int]]:
        if dirpath not in self._mtimes_cache:
            self._mtimes_cache[dirpath] = dir_mtimes(dirpath, client_kwargs=self.client_kwargs)
        return self._mtimes_cache[dirpath]

    def open(self, filepath: str) -> IO[bytes]:
        return s3_open(filepath, client_kwargs=self.client_kwargs)

    def listdir(self, dirpath: str) -> List[str]:
        return list(self.get_dir_mtimes(dirpath).keys())

    def mtime(self, filepath: str) -> Optional[int]:
        dirpath, filename = os.path.split(filepath)
        if dirpath in self._mtimes_cache:
            return self._mtimes_cache[dirpath][filename]
        else:
            return s3_mtime(filepath, client_kwargs=self.client_kwargs)



"""
Module to add xml support
"""
from typing import Any, Dict, List, Optional, Union, cast

import jq
import pandas as pd
import xmltodict

PdDataList = List[Dict[str, Any]]
PdDataDict = Dict[str, List[Any]]


def transform_with_jq(data: Any, jq_filter: str) -> Union[PdDataList, PdDataDict]:
    """Apply a jq filter on data before it's passed to a pd.DataFrame"""
    all_data: Union[List[PdDataList], List[PdDataDict], PdDataList] = jq.all(jq_filter, data)

    # If the data is already presented as a list of rows,
    # then undo the nesting caused by "multiple_output" jq option
    if len(all_data) == 1 and (
        isinstance(all_data[0], list)  # List[PdDataList]
        # detects another valid datastructure [{col1:[value, ...], col2:[value, ...]}]
        or (
            isinstance(all_data[0], dict) and isinstance(list(all_data[0].values())[0], list)
        )  # List[PdDataDict]
    ):
        return all_data[0]
    else:
        return cast(PdDataList, all_data)


def read_xml(
    filepath: str,
    encoding: str = "utf-8",
    filter: Optional[str] = None,
) -> pd.DataFrame:
    data = xmltodict.parse(open(filepath).read(), encoding=encoding)
    if filter is not None:
        data = transform_with_jq(data, filter)
    return pd.DataFrame(data)



"""
This module provides access to the base Fetcher class and its registry
where all the fetchers are saved.
The main purpose of this fetcher is to retrieve a file or a list of files in case of a regex
filepath by calling the right registered subclass (local, ftp...) matching the filepath scheme.
The subclasses are all declared in the `io` directory.
"""
import fnmatch
import os
import re
from abc import ABCMeta, abstractmethod
from enum import Enum
from typing import (
    IO,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Pattern,
    Type,
    TypeVar,
    Union,
)
from urllib.parse import urlparse

from peakina.helpers import mdtm_to_string

F = TypeVar("F", bound=Type["Fetcher"])


class MatchEnum(str, Enum):
    REGEX = "regex"
    GLOB = "glob"


def register(schemes: Union[str, List[str]]) -> Callable[[F], F]:
    if isinstance(schemes, str):
        schemes = [schemes]

    def f(cls: F) -> F:
        for scheme in schemes:
            cls.registry[scheme] = cls
        return cls

    return f


class Fetcher(metaclass=ABCMeta):
    """
    Base class used to:
     - list the files in a directory
     - retrieve the last modification time of a file
     - retrieve one (or many in case of a regex) TextIO from a path
    This class is used by calling `get_fetcher`, which reads the scheme of the path
    ('ftp', 's3', ...) and redirects to the right fetcher.
    All the `Fetcher` subclasses need to implement basic methods in order to be used properly.
    """

    registry: Dict[str, Type["Fetcher"]] = {}

    def __init__(self, **kwargs: Any) -> None:
        self.extra_kwargs = kwargs

    @classmethod
    def get_fetcher(cls, filepath: str, **fetcher_kwargs: Any) -> "Fetcher":
        scheme = urlparse(filepath).scheme
        return cls.registry[scheme](**fetcher_kwargs)

    @abstractmethod
    def listdir(self, dirpath: str) -> List[str]:
        """List all the files in a directory"""

    @abstractmethod
    def open(self, filepath: str) -> Union[IO[bytes], IO[str]]:
        """Same as builtins `open` method in text mode"""

    @abstractmethod
    def mtime(self, filepath: str) -> Optional[int]:
        """Get last modification time of a file"""

    @staticmethod
    def is_matching(filename: str, match: Optional[MatchEnum], pattern: Pattern[str]) -> bool:
        if match is None:
            return bool(filename == pattern.pattern)
        elif match is MatchEnum.GLOB:
            return bool(fnmatch.fnmatch(filename, pattern.pattern))
        else:
            return bool(pattern.match(filename))

    def get_filepath_list(self, filepath: str, match: Optional[MatchEnum] = None) -> List[str]:
        """Methods to retrieve all the pathes to open"""
        if match is None:
            return [filepath]

        dirpath, basename = os.path.split(filepath)
        pattern = re.compile(basename)
        all_filenames = self.listdir(dirpath)
        matching_filenames = [f for f in all_filenames if self.is_matching(f, match, pattern)]
        return [os.path.join(dirpath, f) for f in sorted(matching_filenames)]

    def get_str_mtime(self, filepath: str) -> Optional[str]:
        try:
            mdtime = self.mtime(filepath)
        except (NotImplementedError, KeyError, OSError):
            mdtime = None
        return mdtm_to_string(mdtime) if mdtime else None

    def get_mtime_dict(self, dirpath: str) -> Dict[str, Optional[str]]:
        return {f: self.get_str_mtime(os.path.join(dirpath, f)) for f in self.listdir(dirpath)}


class fetch:
    """class providing shortcuts for some Fetcher operations"""

    def __init__(self, uri: str, **fetcher_kwargs: Any) -> None:
        self.uri = uri
        self.fetcher = Fetcher.get_fetcher(uri, **fetcher_kwargs)

    @property
    def scheme(self) -> str:
        return urlparse(self.uri).scheme

    @property
    def dirpath(self) -> str:
        return os.path.dirname(self.uri)

    @property
    def basename(self) -> str:
        return os.path.basename(self.uri)

    def open(self) -> Union[IO[bytes], IO[str]]:
        return self.fetcher.open(self.uri)

    def get_str_mtime(self) -> Optional[str]:
        return self.fetcher.get_str_mtime(self.uri)

    def get_mtime_dict(self) -> Dict[str, Optional[str]]:
        dirpath, basename = os.path.split(self.uri)
        return self.fetcher.get_mtime_dict(dirpath)



from os import path
from typing import TYPE_CHECKING, Any, Dict, Hashable, Optional

from peakina.cache import Cache
from peakina.datasource import DataSource

if TYPE_CHECKING:
    import pandas as pd


class DataPool:
    def __init__(
        self,
        config: Dict[Hashable, Dict[str, Any]],
        data_sources_dir: str = "",
        cache: Optional[Cache] = None,
    ) -> None:
        self.cache = cache
        self.datasources: Dict[Hashable, DataSource] = {}
        for ds_id, ds_conf in config.items():
            ds = DataSource(**ds_conf)

            # change local path into absolute path
            if ds.scheme == "" and not path.isabs(ds.uri):
                ds.uri = path.join(data_sources_dir, ds.uri)

            self.datasources[ds_id] = ds

    def __contains__(self, item: Hashable) -> bool:
        return item in self.datasources

    def __getitem__(self, item: Hashable) -> "pd.DataFrame":
        return self.datasources[item].get_df(cache=self.cache)

    def __len__(self) -> int:
        return len(self.datasources)



"""
This module provides mainly methods to validate and detect
the type (CSV, Excel...), the encoding and the separator (for CSV) of a file.
It's also where the reference of all supported types is done.
In order to support a new type, you need to create a reader in the `readers`
package and add the MIME types and the name of the method in `SUPPORTED_FILE_TYPES`
The reader needs to take a filepath as first parameter and return a dataframe
"""

import csv
import inspect
import mimetypes
import os
from datetime import datetime
from enum import Enum
from itertools import islice
from typing import Any, Callable, Dict, List, NamedTuple, Optional, cast

import chardet
import pandas as pd

from peakina.readers import (
    csv_meta,
    excel_meta,
    read_csv,
    read_excel,
    read_json,
    read_xml,
)


class TypeInfos(NamedTuple):
    # All the MIME types for a given type of file
    mime_types: List[str]
    # The method to open a given type of file with the `filepath` as first parameter
    # It needs to return a dataframe
    reader: Callable[..., pd.DataFrame]
    # If the default reader has some missing declared kwargs, it's useful
    # to declare them for `validate_kwargs` method
    reader_kwargs: List[str] = []
    metadata_reader: Optional[Callable[..., Dict[str, Any]]] = None


# For files without MIME types, we make fake MIME types based on detected extension
CUSTOM_MIMETYPES = {".parquet": "peakina/parquet"}

EXTRA_PEAKINA_READER_KWARGS = ["preview_offset", "preview_nrows"]

SUPPORTED_FILE_TYPES = {
    "csv": TypeInfos(
        ["text/csv", "text/tab-separated-values"],
        read_csv,
        [],
        csv_meta,
    ),
    "excel": TypeInfos(
        [
            "application/vnd.ms-excel",
            "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
        ],
        read_excel,
        ["encoding", "decimal"],
        excel_meta,
    ),
    "json": TypeInfos(
        ["application/json"],
        read_json,
        ["filter"],  # this option comes from read_json, which @wraps(pd.read_json)
    ),
    "parquet": TypeInfos(["peakina/parquet"], pd.read_parquet),
    "xml": TypeInfos(["application/xml"], read_xml),
}


# FIXME: Right now mypy does not handle properly the Enum API to generate TypeEnum
#        We would like to have something like:
#        > TypeEnum = Enum('TypeEnum', names={v.upper(): v for v in SUPPORTED_TYPES}, type=str)
class TypeEnum(str, Enum):
    CSV = "csv"
    EXCEL = "excel"
    JSON = "json"
    PARQUET = "parquet"
    XML = "xml"


def detect_type(filepath: str, is_regex: bool = False) -> Optional[TypeEnum]:
    """
    Detects the type of a file, which can be a regex or not!
    Can return None in case of generic extension (filepath='...*') with is_regex=True.
    """
    if is_regex:
        filepath = filepath.rstrip("$")
    mimetype, _ = mimetypes.guess_type(filepath)

    # Fallback on custom MIME types
    if mimetype is None:
        _, fileext = os.path.splitext(filepath)
        if fileext in CUSTOM_MIMETYPES:
            mimetype = CUSTOM_MIMETYPES[fileext]

    if is_regex and mimetype is None:  # generic extension with `is_regex=True`
        return None
    try:
        detected_type = [
            type_
            for type_, type_infos in SUPPORTED_FILE_TYPES.items()
            if mimetype in type_infos.mime_types
        ][0]
        return TypeEnum(detected_type)
    except IndexError:
        raise ValueError(
            f"Unsupported mimetype {mimetype!r}. "
            f'Supported types are: {", ".join(map(repr, SUPPORTED_FILE_TYPES))}.'
        )


def bytes_head(filepath: str, n: int) -> bytes:
    """Returns the first `n` lines of a file as a bytes string."""
    with open(filepath, "rb") as f:
        return b"".join(line for line in islice(f, n))


def str_head(filepath: str, n: int, encoding: Optional[str] = None) -> str:
    """Returns the first `n` lines of a file as a string."""
    with open(filepath, encoding=encoding) as f:
        return "".join(line for line in islice(f, n))


def detect_encoding(filepath: str) -> str:
    """Detects the encoding of a file based on its 100 first lines."""
    return cast(str, chardet.detect(bytes_head(filepath, 100))["encoding"])


def validate_encoding(filepath: str, encoding: Optional[str] = None) -> bool:
    """Validates if `encoding` seems ok to read the file based on its 100 first lines."""
    try:
        str_head(filepath, 100, encoding)
        return True
    except UnicodeDecodeError:
        return False


def detect_sep(filepath: str, encoding: Optional[str] = None) -> str:
    """Detect separator of a CSV file based on its 100 first lines"""
    return csv.Sniffer().sniff(str_head(filepath, 100, encoding)).delimiter


def validate_sep(filepath: str, sep: str = ",", encoding: str = "utf-8") -> bool:
    """
    Validates if the `sep` is a right separator of a CSV file
    (i.e. the dataframe has more than one column).
    """
    try:
        # we want an error to be raised if we can't read the first two lines
        # hence the parameter `error_bad_lines` set to `True`
        df = read_csv(filepath, sep=sep, encoding=encoding, nrows=2, error_bad_lines=True)
        return len(df.columns) > 1
    except pd.errors.ParserError:
        return False


def get_reader_allowed_params(t: TypeEnum) -> List[str]:
    reader = SUPPORTED_FILE_TYPES[t].reader
    return [kw for kw in inspect.signature(reader).parameters]


def validate_kwargs(kwargs: Dict[str, Any], t: Optional[TypeEnum]) -> bool:
    """
    Validate that kwargs are at least in one signature of the methods
    Raises an error if it's not the case
    """
    types: List[TypeEnum] = [t] if t is not None else [TypeEnum(t) for t in SUPPORTED_FILE_TYPES]
    allowed_kwargs: List[str] = []
    for t in types:
        allowed_kwargs += get_reader_allowed_params(t)
        # Add extra allowed kwargs
        allowed_kwargs += SUPPORTED_FILE_TYPES[t].reader_kwargs
        allowed_kwargs += EXTRA_PEAKINA_READER_KWARGS
    bad_kwargs = set(kwargs) - set(allowed_kwargs)
    if bad_kwargs:
        raise ValueError(f'Unsupported kwargs: {", ".join(map(repr, bad_kwargs))}')
    return True


def mdtm_to_string(mtime: int) -> str:
    """Convert the last modification date of a file as an iso string"""
    return datetime.utcfromtimestamp(mtime).isoformat() + "Z"


def pd_read(filepath: str, t: str, kwargs: Dict[str, Any]) -> pd.DataFrame:
    return SUPPORTED_FILE_TYPES[t].reader(filepath, **kwargs)


def get_metadata(filepath: str, type: str, reader_kwargs: Dict[str, Any]) -> Dict[str, Any]:
    metadata_reader = SUPPORTED_FILE_TYPES[type].metadata_reader
    return metadata_reader(filepath, reader_kwargs) if metadata_reader else {}



"""
Module to add csv support
"""
from functools import wraps
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

import pandas as pd

if TYPE_CHECKING:
    from os import PathLike

    FilePathOrBuffer = Union[str, bytes, PathLike[str], PathLike[bytes]]

# The chunksize value for previews
PREVIEW_CHUNK_SIZE = 1024


@wraps(pd.read_csv)
def read_csv(
    filepath_or_buffer: "FilePathOrBuffer",
    *,
    # extra `peakina` reader kwargs
    preview_offset: int = 0,
    preview_nrows: Optional[int] = None,
    # change of default values
    keep_default_na: bool = False,  # pandas default: `True`
    error_bad_lines: bool = False,  # pandas default: `True`
    **kwargs: Any,
) -> pd.DataFrame:
    """
    The read_csv method is able to make a preview by reading on chunks
    """
    if preview_nrows is not None or preview_offset:
        chunks = pd.read_csv(
            filepath_or_buffer,
            keep_default_na=keep_default_na,
            error_bad_lines=error_bad_lines,
            **kwargs,
            # keep the first row 0 (as the header) and then skip everything else up to row `preview_offset`
            skiprows=range(1, preview_offset + 1),
            nrows=preview_nrows,
            chunksize=PREVIEW_CHUNK_SIZE,
        )
        return next(chunks)

    return pd.read_csv(
        filepath_or_buffer,
        keep_default_na=keep_default_na,
        error_bad_lines=error_bad_lines,
        **kwargs,
    )


def _line_count(filepath_or_buffer: "FilePathOrBuffer") -> int:
    with open(filepath_or_buffer) as f:
        lines = 0
        buf_size = 1024 * 1024
        read_f = f.read  # loop optimization

        buf = read_f(buf_size)
        while buf:
            lines += buf.count("\n")
            buf = read_f(buf_size)

        return lines


def csv_meta(
    filepath_or_buffer: "FilePathOrBuffer", reader_kwargs: Dict[str, Any]
) -> Dict[str, Any]:
    total_rows = _line_count(filepath_or_buffer)

    if "nrows" in reader_kwargs:
        return {
            "total_rows": total_rows,
            "df_rows": reader_kwargs["nrows"],
        }

    start = 0 + reader_kwargs.get("skiprows", 0)
    end = total_rows - reader_kwargs.get("skipfooter", 0)

    preview_offset = reader_kwargs.get("preview_offset", 0)
    preview_nrows = reader_kwargs.get("preview_nrows", None)

    if preview_nrows is not None:
        return {
            "total_rows": total_rows,
            "df_rows": min(preview_nrows, max(end - start - preview_offset, 0)),
        }
    elif preview_offset:  # and `preview_nrows` is None
        return {
            "total_rows": total_rows,
            "df_rows": max(end - start - preview_offset, 0),
        }
    else:
        return {
            "total_rows": total_rows,
            "df_rows": end - start,
        }



"""
Module to add excel files support
"""
import logging
from functools import wraps
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

import pandas as pd

if TYPE_CHECKING:
    from os import PathLike

    FilePathOrBuffer = Union[str, bytes, PathLike[str], PathLike[bytes]]

LOGGER = logging.getLogger(__name__)


@wraps(pd.read_excel)
def read_excel(
    filepath_or_buffer: "FilePathOrBuffer",
    *,
    # extra `peakina` reader kwargs
    preview_offset: int = 0,
    preview_nrows: Optional[int] = None,
    # change of default values
    keep_default_na: bool = False,  # pandas default: `True`
    **kwargs: Any,
) -> pd.DataFrame:
    df = pd.read_excel(
        filepath_or_buffer,
        keep_default_na=keep_default_na,
        **kwargs,
    )
    # if there are several sheets, pf.read_excel returns a dict {sheet_name: df}
    if isinstance(df, dict):
        for sheet_name, sheet_df in df.items():
            sheet_df["__sheet__"] = sheet_name
        df = pd.concat(df.values(), sort=False)

    if preview_nrows is not None or preview_offset:
        offset = None if preview_nrows is None else preview_offset + preview_nrows
        return df[preview_offset:offset]
    return df


def excel_meta(
    filepath_or_buffer: "FilePathOrBuffer", reader_kwargs: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Returns a dictionary with the meta information of the excel file.
    """
    excel_file = pd.ExcelFile(filepath_or_buffer)
    sheet_names = excel_file.sheet_names

    df = read_excel(excel_file, **reader_kwargs)

    if (sheet_name := reader_kwargs.get("sheet_name", 0)) is None:
        # multiple sheets together
        return {
            "sheetnames": sheet_names,
            "df_rows": df.shape[0],
            "total_rows": sum(excel_file.parse(sheet_name).shape[0] for sheet_name in sheet_names),
        }
    else:
        # single sheet
        return {
            "sheetnames": sheet_names,
            "df_rows": df.shape[0],
            "total_rows": excel_file.parse(sheet_name).shape[0],
        }



"""
This module provides access to the DataSource class.
A datasource is defined by one or many files matching a pattern and some extra parameters like
encoding, separator...and its only method is `get_df` to retrieve the pandas DataFrame for
the given parameters.
"""
import os
from contextlib import suppress
from dataclasses import asdict, field
from datetime import timedelta
from hashlib import md5
from typing import IO, Any, Dict, Generator, Iterable, Optional, Union
from urllib.parse import urlparse, uses_netloc, uses_params, uses_relative

import pandas as pd
from pydantic.dataclasses import dataclass
from slugify import slugify

from peakina.cache import Cache
from peakina.helpers import (
    TypeEnum,
    detect_encoding,
    detect_sep,
    detect_type,
    get_metadata,
    get_reader_allowed_params,
    pd_read,
    validate_encoding,
    validate_kwargs,
    validate_sep,
)
from peakina.io import Fetcher, MatchEnum

AVAILABLE_SCHEMES = set(Fetcher.registry) - {""}  # discard the empty string scheme
PD_VALID_URLS = set(uses_relative + uses_netloc + uses_params) | AVAILABLE_SCHEMES


@dataclass
class DataSource:
    uri: str
    type: Optional[TypeEnum] = None
    match: Optional[MatchEnum] = None
    expire: Optional[timedelta] = None
    reader_kwargs: Dict[str, Any] = field(default_factory=dict)
    fetcher_kwargs: Dict[str, Any] = field(default_factory=dict)

    def __post_init_post_parse__(self) -> None:
        self._fetcher: Optional[Fetcher] = None
        self.scheme = urlparse(self.uri).scheme
        if self.scheme not in PD_VALID_URLS:
            raise AttributeError(f"Invalid scheme {self.scheme!r}")

        self.type = self.type or detect_type(self.uri, is_regex=bool(self.match))

        validate_kwargs(self.reader_kwargs, self.type)

    @property
    def fetcher(self) -> Fetcher:
        if self._fetcher is None:
            self._fetcher = Fetcher.get_fetcher(self.uri, **self.fetcher_kwargs)
        return self._fetcher

    @property
    def hash(self) -> str:
        identifier = asdict(self)
        del identifier["expire"]
        hash_ = md5(str(identifier).encode("utf-8")).hexdigest()
        filename = slugify(os.path.basename(self.uri), separator="_")
        return f"_{filename}_{hash_}"

    def get_metadata(self) -> Dict[str, Any]:
        """Return datasource metadata (e.g. excel sheetnames)"""
        if self.match:
            return {}  # no metadata for matched datasources
        with self.fetcher.open(self.uri) as f:
            assert self.type is not None
            return get_metadata(f.name, self.type, self.reader_kwargs)

    @staticmethod
    def _get_single_df(
        stream: Union[IO[bytes], IO[str]], filetype: Optional[TypeEnum], **kwargs: Any
    ) -> Union[pd.DataFrame, Iterable[pd.DataFrame]]:
        """
        Read a stream and retrieve the data frame or data frame generator (chunks)
        It uses `stream.name`, which is the path to a local file (often temporary)
        to avoid closing it. It will be closed at the end of the method.
        """
        if filetype is None:
            filetype = TypeEnum(detect_type(stream.name))
        allowed_params = get_reader_allowed_params(filetype)

        # Check encoding
        encoding = kwargs.get("encoding", "utf-8")
        if "encoding" in allowed_params:
            if not validate_encoding(stream.name, encoding):
                encoding = detect_encoding(stream.name)
            kwargs["encoding"] = encoding

        # Check separator for CSV files if it's not set
        if "sep" in allowed_params and "sep" not in kwargs:
            if not validate_sep(stream.name, encoding=encoding):
                kwargs["sep"] = detect_sep(stream.name, encoding)

        try:
            df = pd_read(stream.name, filetype, kwargs)
        finally:
            stream.close()

        return df

    def get_matched_datasources(self) -> Generator["DataSource", None, None]:
        my_args = asdict(self)
        for uri in self.fetcher.get_filepath_list(self.uri, self.match):
            overriden_args = {**my_args, "uri": uri, "match": None}
            yield DataSource(**overriden_args)  # type: ignore[arg-type]

    def get_dfs(self, cache: Optional[Cache] = None) -> Generator[pd.DataFrame, None, None]:
        """
        From the conf of the datasource, returns a generator
        with all the dataframes
        The generator can have a single dataframe (single file as input
        without options) or many (e.g. with `match` or `chunksize`)
        """
        by_chunk = self.reader_kwargs.get("chunksize") is not None
        with_cache = cache is not None and self.expire and not by_chunk

        for datasource in self.get_matched_datasources():
            if with_cache:
                cache_key = datasource.hash
                cache_mtime = None
                with suppress(NotImplementedError, KeyError, OSError):
                    cache_mtime = self.fetcher.mtime(datasource.uri)

                with suppress(KeyError):
                    assert cache is not None
                    df = cache.get(key=cache_key, mtime=cache_mtime, expire=self.expire)
                    yield df
                    continue

            stream = self.fetcher.open(datasource.uri)
            try:
                df = self._get_single_df(stream, self.type, **self.reader_kwargs)
                dfs = df if by_chunk else [df]
            except pd.errors.EmptyDataError:
                dfs = [pd.DataFrame()]

            for df in dfs:
                if self.match:
                    df["__filename__"] = os.path.basename(datasource.uri)
                if with_cache:
                    assert cache is not None
                    cache.set(key=cache_key, value=df, mtime=cache_mtime)
                yield df

    def get_df(self, cache: Optional[Cache] = None) -> pd.DataFrame:
        return pd.concat([x for x in self.get_dfs(cache=cache)], sort=False).reset_index(drop=True)


def read_pandas(
    uri: str,
    *,
    type: Optional[TypeEnum] = None,
    match: Optional[MatchEnum] = None,
    expire: Optional[timedelta] = None,
    fetcher_kwargs: Optional[Dict[str, Any]] = None,
    **reader_kwargs: Any,
) -> pd.DataFrame:
    return DataSource(
        uri=uri,
        type=type,
        match=match,
        expire=expire,
        fetcher_kwargs=fetcher_kwargs or {},
        reader_kwargs=reader_kwargs,
    ).get_df()



from .cache import Cache, CacheEnum
from .datapool import DataPool
from .datasource import AVAILABLE_SCHEMES, DataSource, read_pandas
from .helpers import TypeEnum
from .io import MatchEnum, fetch

__all__ = (
    "AVAILABLE_SCHEMES",
    "Cache",
    "CacheEnum",
    "DataPool",
    "DataSource",
    "MatchEnum",
    "TypeEnum",
    "fetch",
    "read_pandas",
)



"""
Module to enhance pandas.read_json with JQ filter
"""
import json
from functools import wraps
from typing import TYPE_CHECKING, Any, Optional, Union

import jq
import pandas as pd

if TYPE_CHECKING:
    from os import PathLike

    FilePathOrBuffer = Union[str, bytes, PathLike[str], PathLike[bytes]]


def transform_with_jq(json_input: str, jq_filter: str) -> str:
    """Apply a jq filter on raw json input, outputs raw json (may be on several lines)"""
    return jq.text(jq_filter, json.loads(json_input))  # type: ignore[no-any-return]


@wraps(pd.read_json)
def read_json(
    path_or_buf: "FilePathOrBuffer",
    encoding: str = "utf-8",
    filter: Optional[str] = None,
    *args: Any,
    **kwargs: Any,
) -> pd.DataFrame:
    if filter is not None:
        with open(path_or_buf, encoding=encoding) as f:
            path_or_buf = transform_with_jq(f.read(), filter)
    return pd.read_json(path_or_buf, encoding=encoding, *args, **kwargs)
